""" Fakts CLI module.

This module contains the CLI for the fakts_next package.
It allows you act and interface with the fakts_next package
from the command line.

Currently it only supports advertising a fakts_next endpoint
on the local network, but in the future it will also
support other features, like fetching the configuration
from a remote endpoint.

"""
